@@ -1,4 +1,6 @@
 import copy
+import itertools
+import warnings
 from types import ModuleType
 from typing import Any, Callable, Dict, List, Optional
 
@@ -8,11 +10,11 @@ from spacy.matcher import Matcher
 from spacy.tokens import Span
 
 from replacy import default_match_hooks
-from replacy.db import get_forms_lookup, get_match_dict, get_match_dict_schema
+from replacy.db import get_forms_lookup, get_match_dict, get_match_dict_schema, load_lm
 from replacy.inflector import Inflector
+from replacy.scorer import KenLMScorer
 from replacy.version import __version__
 
-
 # set known extensions:
 known_string_extensions = ["description", "match_name", "category", "comment"]
 known_list_extensions = ["suggestions"]
@@ -55,7 +57,8 @@ class ReplaceMatcher:
         forms_lookup=None,
         custom_match_hooks: Optional[ModuleType] = None,
         allow_multiple_whitespaces=False,
-        lemmatizer="pyInflect",
+        max_suggestions_count=1000,
+        lm_path=None,
     ):
         self.default_match_hooks = default_match_hooks
         self.custom_match_hooks = custom_match_hooks
@@ -67,8 +70,10 @@ class ReplaceMatcher:
         self.matcher = Matcher(self.nlp.vocab)
         self._init_matcher()
         self.spans: List[Span] = []
-        self.inflector = Inflector(
-            nlp=self.nlp, forms_lookup=self.forms_lookup, lemmatizer=lemmatizer
+        self.inflector = Inflector(nlp=self.nlp, forms_lookup=self.forms_lookup)
+        self.max_suggestions_count = max_suggestions_count
+        self.scorer = (
+            KenLMScorer(nlp=self.nlp, model=load_lm(lm_path)) if lm_path else None
         )
 
         # set custom extensions for any unexpected keys found in the match_dict
@@ -131,7 +136,7 @@ class ReplaceMatcher:
                     pred = template(**kwargs)
             elif type(args) == dict:
                 # should we force them to use kwargs?
-                print(
+                warnings.warn(
                     f"WARNING: dict passed as sole args argument. Calling {hook['name']} "
                     f"with single argument {args}. If you want to call with keyword arguments, use kwargs"
                 )
@@ -149,64 +154,156 @@ class ReplaceMatcher:
             predicates.append(pred)
         return predicates
 
+    def get_item(self, item, doc, start, end):
+
+        item_options = []
+        # set
+        if "TEXT" in item:
+            if isinstance(item["TEXT"], dict):
+                item_options = item["TEXT"].get("IN", [])
+            elif isinstance(item["TEXT"], str):
+                item_options = [item["TEXT"]]
+        # copy
+        elif "PATTERN_REF" in item:
+            ref = int(item["PATTERN_REF"])
+            if ref >= 0:
+                refd_token = doc[start + ref]
+            else:
+                # this is confusing. Example:
+                # doc = nlp("I like apples, blood oranges, and bananas")
+                # start = 2, end = 9 gives doc[start:end] == "apples, blood oranges, and bananas"
+                # but doc[9] != "bananas", it is an IndexError, the last token is end-1
+                # so, per python conventions, PATTERN_REF = -1 would mean the last matched token
+                # so we can just add ref and end if ref is negative
+                # to do: match again to get multi-token
+                refd_token = doc[end + ref]
+            item_options = [refd_token.text]
+
+        return item_options
+
+    def inflect_item(self, item_options, item, doc, start, end, match_name):
+        # set
+        if "INFLECTION" in item:
+            inflection_value = item["INFLECTION"]
+            inflection_type = Inflector.get_inflection_type(inflection_value)
+            if inflection_type == "pos":
+                # set by pos
+                item_options = (
+                    seq(item_options)
+                    .map(
+                        lambda x: self.inflector.inflect_or_lookup(
+                            x, pos=inflection_value
+                        )
+                    )
+                    .flatten()
+                    .list()
+                )
+            elif inflection_type == "tag":
+                # set by tag
+                item_options = (
+                    seq(item_options)
+                    .map(
+                        lambda x: self.inflector.inflect_or_lookup(
+                            x, tag=inflection_value
+                        )
+                    )
+                    .flatten()
+                    .list()
+                )
+            else:
+                # get all forms
+                item_options = (
+                    seq(item_options)
+                    .map(lambda x: self.inflector.inflect_or_lookup(x, pos=None))
+                    .flatten()
+                    .list()
+                )
+        # copy
+        elif "FROM_TEMPLATE_ID" in item:
+            template_id = int(item["FROM_TEMPLATE_ID"])
+            index = None
+            for i, token in enumerate(self.match_dict[match_name]["patterns"]):
+                if "TEMPLATE_ID" in token and token["TEMPLATE_ID"] == template_id:
+                    index = i
+                    break
+            if index is not None:
+                item_options = (
+                    seq(item_options)
+                    .map(lambda x: self.inflector.auto_inflect(doc, x, start + index))
+                    .flatten()
+                    .list()
+                )
+        return item_options
+
+    def case_item(self, item_options, item):
+        # This should probably be a list of ops
+        # and we should have a parser class
+        if "REPLACY_OP" in item:
+            op = item["REPLACY_OP"]
+            if op == "LOWER":
+                item_options = [t.lower() for t in item_options]
+            if op == "TITLE":
+                item_options = [t.title() for t in item_options]
+            if op == "UPPER":
+                item_options = [t.upper() for t in item_options]
+        return item_options
+
     def process_suggestions(self, pre_suggestion, doc, start, end, match_name):
         """
-        Perform inflection and replace references to the matched token
-        example of pattern and pre_suggestion
-        pattern: "LEMMA": "chock", "TEMPLATE_ID": 1
-        pre_suggestion: "TEXT": "chalk", "FROM_TEMPLATE_ID": 1
-        inflect suggestion "chalk" according to form of "cholk" from patterns
+        Suggestion text:
+            - set: "TEXT": "cat"
+            - choose one from: "TEXT": {"IN": ["a", "b"]}
+            - copy from pattern: "PATTERN_REF": 3 (copy from 3rd pattern match)
+        Set suggestion text inflection:
+            - set by tag: "INFLECTION": "VBG" (returns one)
+            - set by pos: "INFLECTION": "NOUN" (returns many. ex. NNS, NN)
+            - get all: "INFLECTION": "ALL" (returns a lot, use infrequently)
+            - copy from pattern: "FROM_TEMPLATE_ID": 2 (copy from token with "TEMPLATE_ID":2)
+        Suggestions case matching:
+            - lowercase: "REPLACY_OP: "LOWER"
+            - title: "REPLACY_OP: "TITLE"
+            - upper: "REPLACY_OP: "UPPER"       
         """
-        text_list = []
+        options = []
         for item in pre_suggestion:
-            try:
-                text = item["TEXT"]
-            except KeyError:
-                ref = int(item["PATTERN_REF"])
-                if ref >= 0:
-                    refd_token = doc[start + ref]
-                else:
-                    # this is confusing. Example:
-                    # doc = nlp("I like apples, blood oranges, and bananas")
-                    # start = 2, end = 9 gives doc[start:end] == "apples, blood oranges, and bananas"
-                    # but doc[9] != "bananas", it is an IndexError, the last token is end-1
-                    # so, per python conventions, PATTERN_REF = -1 would mean the last matched token
-                    # so we can just add ref and end if ref is negative
-                    refd_token = doc[end + ref]
-            changed_text = None
 
-            # check if inflect
-            if "FROM_TEMPLATE_ID" in item:
-                template_id = item["FROM_TEMPLATE_ID"]
-                index = None
-                for i, token in enumerate(self.match_dict[match_name]["patterns"]):
-                    if "TEMPLATE_ID" in token and token["TEMPLATE_ID"] == template_id:
-                        index = i
-                        break
-                if index is not None:
-                    changed_text = self.inflector.auto_inflect(doc, text, start + index)
-            elif "PATTERN_REF" in item:
-                if "INFLECTION" in item:
-                    form = item["INFLECTION"]
-                    changed_text = self.inflector.inflect_or_lookup(refd_token, form)
-                else:
-                    changed_text = refd_token.text
+            item_options = self.get_item(item, doc, start, end)
+            inflected_options = self.inflect_item(
+                item_options, item, doc, start, end, match_name
+            )
+            cased_options = self.case_item(inflected_options, item)
+            options.append(cased_options)
 
-                # This should probably be a list of ops
-                # and we should have a parser class
-                if "REPLACY_OP" in item:
-                    op = item["REPLACY_OP"]
-                    if op == "LOWER":
-                        changed_text = changed_text.lower()
-                    if op == "TITLE":
-                        changed_text = changed_text.title()
-                    if op == "UPPER":
-                        changed_text = changed_text.upper()
-            if changed_text:
-                text_list.append(changed_text)
-            elif len(text):
-                text_list.append(text)
-        return " ".join(text_list)
+        # assert there aren't more than max_suggestions_count
+        # otherwise raise warning and return []
+        suggestions_count = (
+            seq(options).map(lambda x: len(x)).reduce(lambda x, y: x * y)
+        )
+
+        if suggestions_count > self.max_suggestions_count:
+            warnings.warn(
+                f"Got {suggestions_count} suggestions, max is {self.max_suggestions_count}. \
+                Will fallback to empty suggestions."
+            )
+            opt_text = []
+        else:
+            opt_combinations = list(itertools.product(*options))
+            opt_text = [" ".join(list(o)) for o in opt_combinations]
+
+        return opt_text
+
+    def score_suggestion(self, doc, span, suggestion):
+        text = " ".join([doc[: span.start].text, suggestion, doc[span.end :].text])
+        return self.scorer(text)
+
+    def sort_suggestions(self, doc, spans):
+        for span in spans:
+            if len(span._.suggestions) > 1:
+                span._.suggestions = sorted(
+                    span._.suggestions,
+                    key=lambda x: self.score_suggestion(doc, span, x),
+                )
+        return spans
 
     def get_callback(self, match_name, match_hooks):
         """
@@ -233,10 +330,12 @@ class ReplaceMatcher:
 
             pre_suggestions = self.match_dict[match_name]["suggestions"]
 
-            span._.suggestions = [
-                self.process_suggestions(x, doc, start, end, match_name)
-                for x in pre_suggestions
-            ]
+            span._.suggestions = []
+
+            for x in pre_suggestions:
+                span._.suggestions += self.process_suggestions(
+                    x, doc, start, end, match_name
+                )
 
             span._.description = self.match_dict[match_name].get("description", "")
             span._.category = self.match_dict[match_name].get("category", "")
@@ -290,4 +389,7 @@ class ReplaceMatcher:
         # this fills up self.spans
         matches = self.matcher(sent)
 
+        if self.scorer:
+            self.spans = self.sort_suggestions(sent, self.spans)
+
         return self.spans
